Require Export auxiliary_prop.
Require Import List.

(* subTp' derivations imply well-formed typing environments *)

Lemma subTp'_ensures_okEnv:
      forall G S T,
      subTp' G S T -> okEnv G.
induction 1.
assumption. assumption. assumption. assumption. assumption.
Qed.

(* Auxiliary lemmas: from deep to shallow *)

Lemma isinG2envBook: forall G: envTp, forall x: Var, forall U: Tp, isinG x U G -> (book2Prop (envTp2envBook G)) -> (envBook x U).

Proof.

induction G; intros.
 inversion_clear H.

 inversion H.
 inversion_clear H2.
  inversion_clear H4.
  rewrite <- H1 in H0.
  simpl in H0.
  rewrite H2; rewrite H5; tauto.

  apply IHG; auto.
  rewrite <- H1 in H0; simpl in H0; tauto.

Qed.

Lemma Gclosed2closed: forall S: Tp, forall G: envTp, Gclosed S G -> (book2Prop (envTp2envBook G)) -> closed S.

Proof.

induction S; unfold Gclosed; unfold closed; intros.
 inversion_clear H1.

 inversion H1.
 assert (exists U : Tp, isinG x U G).
  apply H; auto.

  inversion_clear H2.
  exists x0; apply isinG2envBook with G; auto.
  rewrite <- H3; assumption.

 assert (exists U : Tp, isinG x U G).
  apply H; assumption.

  inversion_clear H2.
  exists x0.
  inversion_clear H1.
  inversion_clear H2.
   apply isinG2envBook with G; auto.

   apply isinG2envBook with G; auto.

 assert (exists U : Tp, isinG x U G).
  apply H0; assumption.

  inversion_clear H3.
  exists x0.
  apply isinG2envBook with G; assumption.

Qed.

Lemma freshList: forall G: envTp, forall x: Var, (notin_list x (envTp2list G)) -> Gfresh x G.

Proof.

induction G; intros.
 apply GfVoid.

 generalize H.
 clear H.
 case a; intros.
 inversion H.
 apply GfGrow.
  apply IHG; auto.

  assumption.

Qed.

Lemma domGtoT_notin: forall G: envTp, forall x: Var,
                     notin x (domGtoT G) -> Gfresh x G.
induction G; intros.
 apply GfVoid.
generalize H.
  clear H.
  elim a; intros.
  apply GfGrow.
 apply IHG; auto.
   simpl in H.
   inversion H; assumption.
elim (LEM_Var x a0); intros.
 rewrite H0 in H; inversion_clear H.
   inversion_clear H1.
    absurd (a0 = a0); auto.
assumption.
Qed.

Lemma unsatG: forall G: envTp,
              exists x:Var, Gfresh x G.
intro; elim (unsat (domGtoT G)); intros; exists x; apply domGtoT_notin; assumption.
Qed.

Lemma Gclosed_lemma: forall G: envTp, forall S T: Tp,
                     subTp' G S T ->
                     Gclosed S G /\ Gclosed T G.
intros. assert (okEnv G).
apply subTp'_ensures_okEnv with S T. assumption.
generalize H0 H; clear H0 H.
generalize G S T; clear G S T.

induction 2; intros.
 split.
  assumption.
 unfold Gclosed in |- *; intros.
   inversion_clear H2.
unfold Gclosed in |- *; split; intros; inversion H2; exists U; assumption.
split.
 unfold Gclosed in |- *; intros.
   inversion H2.
   exists U; assumption.
 tauto.
unfold Gclosed in |- *; split; intros.
 inversion H2.
   inversion H4.
  cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | auto ].
    inversion_clear H7.
    unfold Gclosed in H9.
    apply H9; auto.
 cut (Gclosed S2 G /\ Gclosed T2 G); [ intro | auto ].
   inversion_clear H7.
   unfold Gclosed in H8; apply H8; auto.
inversion H2.
  inversion H4.
 cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | auto ].
   inversion_clear H7.
   unfold Gclosed in H8; apply H8; auto.
cut (Gclosed S2 G /\ Gclosed T2 G); [ intro | auto ].
  inversion_clear H7.
  unfold Gclosed in H9; apply H9; auto.

unfold Gclosed in |- *; split; intros.

 inversion H3.
   inversion H5.
  cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | auto ].
    inversion_clear H8.
    unfold Gclosed in H10; apply H10; auto.
 elim (unsatG ((x, T1) :: G)); intros.
   cut (Gclosed (S2 x0) ((x0, T1) :: G) /\ Gclosed (T2 x0) ((x0, T1) :: G));
    [ intro | apply H2; auto ].
  inversion_clear H9.
    unfold Gclosed in H10.
    cut (exists U : Tp, isinG x U ((x0, T1) :: G));
     [ intro | apply H10; auto ].
   inversion H9.
     inversion H12.
     inversion H14.
    inversion_clear H17.
      inversion H8.
       absurd (x = x0).
     auto.
    auto.
   exists x1; assumption.
  apply H7; auto.
    inversion H8.
    auto.
 apply okGrow.
  auto.
 inversion H8; auto.
 cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | apply IHsubTp'; assumption ].
    tauto.
 apply okGrow; auto.
  inversion H8; auto.
 cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | apply IHsubTp'; assumption ].
    tauto.
 inversion H3.
   inversion H5.
  cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | auto ].
    inversion_clear H8.
    unfold Gclosed in H9; apply H9; auto.
 elim (unsatG ((x, T1) :: G)); intros.
   cut (Gclosed (S2 x0) ((x0, T1) :: G) /\ Gclosed (T2 x0) ((x0, T1) :: G));
    [ intro | apply H2; auto ].
  inversion_clear H9.
    unfold Gclosed in H11.
    cut (exists U : Tp, isinG x U ((x0, T1) :: G));
     [ intro | apply H11; auto ].
   inversion H9.
     inversion H12.
     inversion H14.
    inversion_clear H17.
      inversion H8.
       absurd (x = x0).
     auto.
    auto.
   exists x1; assumption.
  apply H7; auto.
    inversion H8.
    auto.
 apply okGrow.
  auto.
 inversion H8; auto.
 cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | apply IHsubTp'; assumption ].
    tauto.
 apply okGrow; auto.
  inversion H8; auto.
 cut (Gclosed T1 G /\ Gclosed S1 G); [ intro | apply IHsubTp'; assumption ].
    tauto.
Qed.

(* Internal Adequacy: from deep to shallow *)

Lemma exp2imp: forall G: envTp, forall S T: Tp, subTp' G S T -> (book2Prop (envTp2envBook G)) -> subTp S T.

Proof.

induction 1; intros.

apply sub_top; auto.
apply Gclosed2closed with G; auto.

apply sub_var with U; auto.
apply isinG2envBook with G; auto.

apply sub_trs with U; auto.
apply isinG2envBook with G; auto.

apply sub_arr; auto.

 apply sub_fa with (envTp2list G); intros.
  apply IHsubTp'; auto.

  apply H1; auto.
   apply okGrow.
    apply subTp'_ensures_okEnv with T1 S1; auto.

    apply freshList; auto.

    assert (Gclosed T1 G /\ Gclosed S1 G).
     apply Gclosed_lemma; auto.

     tauto.

   simpl.
   tauto.

Qed.